#ifndef DLPLAN_SRC_UTILS_MATH_H
#define DLPLAN_SRC_UTILS_MATH_H


namespace dlplan::utils {

// https://stackoverflow.com/questions/55421835/c-binomial-coefficient-is-too-slow
int binomial_coefficient(int n, int k);

}

#endif
